#
# librefe.rb
#
#   Copyright (c) 2001 Minero Aoki <aamine@cd.xdsl.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU General Public License version 2 or later.
#

class Refe

  class SearchError < StandardError; end


  INDEX_FILE = 'index.refe'

  unless defined? DATA_DIR then
    DATA_DIR = File.expand_path('~/share')
  end

  def self.load
    File.open( "#{DATA_DIR}/#{INDEX_FILE}" ) do |f|
      return Marshal.load(f)
    end
  end


  def initialize( tab )
    @table = tab
    reduce_table tab

    @inv_table = invert_table( tab )

    @c_abbr_table = create_abbr_table( class_names )
    @m_abbr_table = create_abbr_table( @inv_table.keys )
  end

  def reduce_table( table )
    del = {}
    nomethod = {}
    table.each do |k,mt|
      if mt.empty? then
        k = k[0..-2]
        if del.key? k then
          nomethod[k] = true
        end
        del[k] = true
      end
    end
    @no_method_classes = nomethod

    table.delete_if do |k,mt|
      mt.empty?
    end
  end


  def invert_table( table )
    it = {}
    table.each do |cname, methods|
      methods.each_key do |name|
        (it[name] ||= []).push cname
      end
    end

    it
  end

  def create_abbr_table( cand )
    table = {}
    cand.each do |orig|
      str = orig.downcase
      begin
        (table[str] ||= []).push orig
        str.chop!
      end until str.empty?
    end

    table
  end


  def print_classes
    packed_print class_names
  end

  def class_names
    @table.keys + @no_method_classes.keys
  end


  def lookup( klass, method, type )
    ccand = select_class( klass, type )
    ccand.empty? and notfound 'class', klass

    mcand = select_method( method )
    mcand.empty? and notfound 'method', method

    ent = ccand.collect {|c|
        t = mcand.find_all {|m| @table[c][m] }
        t.empty? ? nil : [c,t]
    }.compact

    ent.empty? and notfound 'method', "#{klass}#{type ? type : ' '}#{method}"
    if ent.size != 1 then
      packed_print ent.collect {|c,t| t.collect {|m| c + m } }.flatten
      return
    end
    c, mnames = ent.first
    if mnames.size != 1 then
      packed_print mnames.collect {|m| c + m }
      return
    end
    print_desc c, mnames.first
  end


  def lookup_class( klass, type )
    ccand = select_class( klass, type, true )
    ccand.empty? and notfound 'class', klass

    if ccand.size == 1 and /[\.\#]/ === ccand.first[-1,1] then
      print_class ccand.first
    elsif ccand.size == 2 and ccand[0].chop == ccand[1].chop then
      print_class ccand[0]
      print_class ccand[1]
    else
      packed_print ccand
    end
  end


  def lookup_method( method )
    mcand = select_method( method )
    mcand.empty? and notfound 'method', method

    ent = mcand.collect {|m|
        @inv_table[m].collect {|c| m == method ? c : c + m }
    }.flatten

    if ent.size == 1 then
      m = mcand.first
      c, = @inv_table[m]
      print_desc c, m
    else
      packed_print ent
    end
  end


  def select_class( name, type, usenom = false )
    if type then
      select_class_0( name, type )
    else
      select_class_0( name, '#' ) +
      select_class_0( name, '.' ) +
      if usenom then
        t = @c_abbr_table[name.downcase] and
                t.find_all {|c| not /[\.\#]/ === c[-1,1] }
      end.to_a
    end
  end

  def select_class_0( name, type )
    @table.key? name + type and return [name + type]
    ccand = @c_abbr_table[name.downcase] or return []
    ccand.collect {|c| c[-1,1] == type ? c : nil }.compact
  end

  def select_method( name )
    @inv_table[name] and return [name]
    @m_abbr_table[name.downcase] || []
  end


  def print_class( name )
    print '---- ', name, " ----\n"
    packed_print @table[name].keys
  end

  def print_desc( klass, method )
    puts "#{klass}#{method}"
    puts @table[ klass ][ method ].strip
  end

  def packed_print( arr )
    max = 60
    len = 0
    arr.sort.each do |i|
      if len + 1 + i.size > max then
        puts; len = 0
      end
      print i, ' '; len += i.size
    end
    puts
  end


  def notfound( label, name )
    raise SearchError, "no such #{label}: #{name}"
  end

end
